import pytest
import numpy as np
from numpy.testing import assert_allclose
from scipy.stats import norm, wishart, invgamma, invwishart

import tensorflow as tf

from brainiak.matnormal.covs import (
    CovIdentity,
    CovAR1,
    CovIsotropic,
    CovDiagonal,
    CovDiagonalGammaPrior,
    CovUnconstrainedCholesky,
    CovUnconstrainedCholeskyWishartReg,
    CovUnconstrainedInvCholesky,
    CovKroneckerFactored,
)

# X is m x n, so A sould be m x p

m = 8
n = 4
p = 3

rtol = 1e-7
atol = 1e-7


def logdet_sinv_np(X, sigma):
    # logdet
    sign, logdet = np.linalg.slogdet(sigma)
    logdet_np = sign * logdet
    # sigma-inv
    sinv_np = np.linalg.inv(sigma)
    # solve
    sinvx_np = np.linalg.solve(sigma, X)
    return logdet_np, sinv_np, sinvx_np


def logdet_sinv_np_mask(X, sigma, mask):
    mask_indices = np.nonzero(mask)[0]
    # logdet
    _, logdet_np = np.linalg.slogdet(sigma[np.ix_(mask_indices, mask_indices)])
    # sigma-inv
    sinv_np_ = np.linalg.inv(sigma[np.ix_(mask_indices, mask_indices)])
    # sigma-inverse *
    sinvx_np_ = sinv_np_.dot(X[mask_indices, :])

    sinv_np = np.zeros_like(sigma)
    sinv_np[np.ix_(mask_indices, mask_indices)] = sinv_np_
    sinvx_np = np.zeros_like(X)
    sinvx_np[mask_indices, :] = sinvx_np_

    return logdet_np, sinv_np, sinvx_np


X = norm.rvs(size=(m, n))
X_tf = tf.constant(X)
A = norm.rvs(size=(m, p))
A_tf = tf.constant(A)
eye = tf.eye(m, dtype=tf.float64)


def test_CovConstant(seeded_rng):

    cov_np = wishart.rvs(df=m + 2, scale=np.eye(m))
    cov = CovUnconstrainedCholesky(Sigma=cov_np)

    # verify what we pass is what we get
    cov_tf = cov._cov
    assert_allclose(cov_tf, cov_np)

    # compute the naive version
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_CovIdentity(seeded_rng):

    cov = CovIdentity(size=m)

    # compute the naive version
    cov_np = np.eye(m)
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_CovIsotropic(seeded_rng):

    cov = CovIsotropic(size=m)

    # compute the naive version
    cov_np = cov._cov * np.eye(cov.size)
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)

    # test initialization
    cov = CovIsotropic(var=0.123, size=3)
    assert_allclose(np.exp(cov.log_var.numpy()), 0.123)


def test_CovDiagonal(seeded_rng):

    cov = CovDiagonal(size=m)

    # compute the naive version
    cov_np = cov._cov
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_CovDiagonal_initialized(seeded_rng):

    cov_np = np.diag(np.exp(np.random.normal(size=m)))
    cov = CovDiagonal(size=m, diag_var=np.diag(cov_np))

    # compute the naive version
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_CovDiagonalGammaPrior(seeded_rng):

    cov_np = np.diag(np.exp(np.random.normal(size=m)))
    cov = CovDiagonalGammaPrior(size=m, sigma=np.diag(cov_np), alpha=1.5,
                                beta=1e-10)

    ig = invgamma(1.5, scale=1e-10)

    # compute the naive version
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    penalty_np = np.sum(ig.logpdf(1 / np.diag(cov_np)))
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)
    assert_allclose(penalty_np, cov.logp, rtol=rtol)


def test_CovUnconstrainedCholesky(seeded_rng):

    cov = CovUnconstrainedCholesky(size=m)

    L = cov.L.numpy()
    cov_np = L @ L.T
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_CovUnconstrainedCholeskyWishartReg(seeded_rng):

    cov = CovUnconstrainedCholeskyWishartReg(size=m)

    L = cov.L.numpy()
    cov_np = L @ L.T

    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)
    # now compute the regularizer
    reg = wishart.logpdf(cov_np, df=m + 2, scale=1e10 * np.eye(m))
    assert_allclose(reg, cov.logp, rtol=rtol)


def test_CovUnconstrainedInvCholesky(seeded_rng):

    init = invwishart.rvs(scale=np.eye(m), df=m + 2)
    cov = CovUnconstrainedInvCholesky(invSigma=init)

    Linv = cov.Linv
    L = np.linalg.inv(Linv)
    cov_np = L @ L.T

    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_Cov2FactorKron(seeded_rng):
    assert m % 2 == 0
    dim1 = int(m / 2)
    dim2 = 2

    with pytest.raises(TypeError) as excinfo:
        cov = CovKroneckerFactored(sizes=dim1)
    assert "sizes is not a list" in str(excinfo.value)

    cov = CovKroneckerFactored(sizes=[dim1, dim2])

    L1 = (cov.L[0]).numpy()
    L2 = (cov.L[1]).numpy()
    cov_np = np.kron(np.dot(L1, L1.transpose()), np.dot(L2, L2.transpose()))
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)

    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_Cov3FactorKron(seeded_rng):
    assert m % 4 == 0
    dim1 = int(m / 4)
    dim2 = 2
    dim3 = 2
    cov = CovKroneckerFactored(sizes=[dim1, dim2, dim3])

    L1 = (cov.L[0]).numpy()
    L2 = (cov.L[1]).numpy()
    L3 = (cov.L[2]).numpy()
    cov_np = np.kron(
        np.kron(np.dot(L1, L1.transpose()), np.dot(L2, L2.transpose())),
        np.dot(L3, L3.transpose()),
    )
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)

    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinv_np, cov.solve(eye), rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_Cov3FactorMaskedKron(seeded_rng):
    assert m % 4 == 0
    dim1 = int(m / 4)
    dim2 = 2
    dim3 = 2

    mask = np.random.binomial(1, 0.5, m).astype(np.int32)

    if sum(mask == 0):
        mask[0] = 1
    mask_indices = np.nonzero(mask)[0]

    cov = CovKroneckerFactored(sizes=[dim1, dim2, dim3], mask=mask)

    L1 = (cov.L[0]).numpy()
    L2 = (cov.L[1]).numpy()
    L3 = (cov.L[2]).numpy()
    cov_np_factor = np.kron(L1, np.kron(L2, L3))[
                            np.ix_(mask_indices, mask_indices)]
    cov_np = np.dot(cov_np_factor, cov_np_factor.transpose())
    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X[mask_indices, :], cov_np)

    assert_allclose(logdet_np, cov.logdet, rtol=rtol, atol=atol)
    assert_allclose(
        sinv_np,
        cov.solve(eye).numpy()[np.ix_(mask_indices, mask_indices)],
        rtol=rtol,
        atol=atol,
    )
    assert_allclose(
        sinvx_np, cov.solve(X_tf).numpy()[
                            mask_indices, :], rtol=rtol, atol=atol
    )


def test_CovAR1(seeded_rng):

    cov = CovAR1(size=m)

    cov_np = np.linalg.inv(cov.solve(eye))

    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)

    # test initialization
    cov = CovAR1(rho=0.3, sigma=1.3, size=3)
    assert_allclose(np.exp(cov.log_sigma.numpy()), 1.3)
    assert_allclose((2 * tf.sigmoid(cov.rho_unc) - 1).numpy(), 0.3)


def test_CovAR1_scan_onsets(seeded_rng):

    cov = CovAR1(size=m, scan_onsets=[0, m // 2])

    # compute the naive version
    cov_np = np.linalg.inv(cov.solve(eye))

    logdet_np, sinv_np, sinvx_np = logdet_sinv_np(X, cov_np)
    assert_allclose(logdet_np, cov.logdet, rtol=rtol)
    assert_allclose(sinvx_np, cov.solve(X_tf), rtol=rtol)


def test_raises(seeded_rng):

    with pytest.raises(RuntimeError):
        CovUnconstrainedCholesky(Sigma=np.eye(3), size=4)

    with pytest.raises(RuntimeError):
        CovUnconstrainedCholesky()

    with pytest.raises(RuntimeError):
        CovUnconstrainedInvCholesky(invSigma=np.eye(3), size=4)

    with pytest.raises(RuntimeError):
        CovUnconstrainedInvCholesky()



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Full Correlation Matrix Analysis (FCMA)

Correlation-based voxel selection
"""

# Authors: Yida Wang
# (Intel Labs), 2016

import numpy as np
import time
from mpi4py import MPI
from scipy.stats.mstats import zscore
from sklearn import model_selection
import sklearn
from . import fcma_extension  # type: ignore
from . import cython_blas as blas  # type: ignore
from ..utils.utils import usable_cpu_count
import logging
import multiprocessing

logger = logging.getLogger(__name__)

__all__ = [
    "VoxelSelector",
]


def _cross_validation_for_one_voxel(clf, vid, num_folds, subject_data, labels):
    """Score classifier on data using cross validation."""
    # no shuffling in cv
    skf = model_selection.StratifiedKFold(n_splits=num_folds,
                                          shuffle=False)
    scores = model_selection.cross_val_score(clf, subject_data,
                                             y=labels,
                                             cv=skf, n_jobs=1)
    logger.debug(
        'cross validation for voxel %d is done' %
        vid
    )
    return (vid, scores.mean())


class VoxelSelector:
    """Correlation-based voxel selection component of FCMA.

    Parameters
    ----------

    labels: list of 1D array
        the condition labels of the epochs
        len(labels) labels equals the number of epochs

    epochs_per_subj: int
        The number of epochs of each subject

    num_folds: int
        The number of folds to be conducted in the cross validation

    raw_data: list of 2D array in shape [epoch length, nVoxels]
        Assumption: 1. all activity data contains the same number of voxels
                    2. the activity data has been z-scored,
                       ready to compute correlation as matrix multiplication
                    3. all subjects have the same number of epochs
                    4. epochs belonging to the same subject are adjacent
                       in the list
                    5. if MPI jobs are running on multiple nodes, the path
                       used must be on a filesystem shared by all nodes

    raw_data2: Optional, list of 2D array in shape [epoch length, nVoxels]
        raw_data2 shares the data structure of the assumptions of raw_data
        If raw_data2 is None, the correlation will be computed as
        raw_data by raw_data.
        If raw_data2 is specified, len(raw_data) MUST equal len(raw_data2),
        the correlation will be computed as raw_data by raw_data2.

    voxel_unit: int, default 64
        The number of voxels assigned to a worker each time

    process_num: Optional[int]
        The maximum number of processes used in cross validation.
        If None, the number of processes will equal
        the number of available hardware threads, considering cpusets
        restrictions.
        If 0, cross validation will not use python multiprocessing.

    master_rank: int, default 0
        The process which serves as the master
    """
    def __init__(self,
                 labels,
                 epochs_per_subj,
                 num_folds,
                 raw_data,
                 raw_data2=None,
                 voxel_unit=64,
                 process_num=4,
                 master_rank=0):
        self.labels = labels
        self.epochs_per_subj = epochs_per_subj
        self.num_folds = num_folds
        self.raw_data = raw_data
        self.num_voxels = raw_data[0].shape[1]
        self.raw_data2 = raw_data2
        self.num_voxels2 = raw_data2[0].shape[1] \
            if raw_data2 is not None else self.num_voxels
        self.voxel_unit = voxel_unit
        usable_cpus = usable_cpu_count()
        if process_num is None:
            self.process_num = usable_cpus
        else:
            self.process_num = np.min((process_num, usable_cpus))
        if self.process_num == 0:
            self.use_multiprocessing = False
        else:
            self.use_multiprocessing = True
        self.master_rank = master_rank
        if self.raw_data2 is not None \
                and len(self.raw_data) != len(self.raw_data2):
            raise ValueError('The raw data lists must have the same number '
                             'of elements for computing the correlations '
                             'element by element')
        if self.num_voxels == 0 or self.num_voxels2 == 0:
            raise ValueError('Zero processed voxels')
        if MPI.COMM_WORLD.Get_size() == 1:
            raise RuntimeError('one process cannot run the '
                               'master-worker model')
        if self.master_rank >= MPI.COMM_WORLD.Get_size():
            logger.warn('Master rank exceeds the number of '
                        'launched processes, set to 0')
            self.master_rank = 0

    # tags for MPI messages
    _WORKTAG = 0
    _TERMINATETAG = 1

    def run(self, clf):
        """Run correlation-based voxel selection in master-worker model.

        Sort the voxels based on the cross-validation accuracy
        of their correlation vectors

        Parameters
        ----------
        clf: classification function
            the classifier to be used in cross validation

        Returns
        -------
        results: list of tuple (voxel_id, accuracy)
            the accuracy numbers of all voxels, in accuracy descending order
            the length of array equals the number of voxels
        """
        rank = MPI.COMM_WORLD.Get_rank()
        if rank == self.master_rank:
            results = self._master()
            # Sort the voxels
            results.sort(key=lambda tup: tup[1], reverse=True)
        else:
            self._worker(clf)
            results = []
        return results

    def _master(self):
        """Master node's operation.

        Assigning tasks to workers and collecting results from them

        Parameters
        ----------
        None

        Returns
        -------
        results: list of tuple (voxel_id, accuracy)
            the accuracy numbers of all voxels, in accuracy descending order
            the length of array equals the number of voxels
        """
        logger.info(
            'Master at rank %d starts to allocate tasks',
            MPI.COMM_WORLD.Get_rank()
        )
        results = []
        comm = MPI.COMM_WORLD
        size = comm.Get_size()
        sending_voxels = self.voxel_unit if self.voxel_unit < self.num_voxels \
            else self.num_voxels
        current_task = (0, sending_voxels)
        status = MPI.Status()
        # using_size is used when the number of tasks
        # is smaller than the number of workers
        using_size = size
        for i in range(0, size):
            if i == self.master_rank:
                continue
            if current_task[1] == 0:
                using_size = i
                break
            logger.debug(
                'master starts to send a task to worker %d' %
                i
            )
            comm.send(current_task,
                      dest=i,
                      tag=self._WORKTAG)
            next_start = current_task[0] + current_task[1]
            sending_voxels = self.voxel_unit \
                if self.voxel_unit < self.num_voxels - next_start \
                else self.num_voxels - next_start
            current_task = (next_start, sending_voxels)

        while using_size == size:
            if current_task[1] == 0:
                break
            result = comm.recv(source=MPI.ANY_SOURCE,
                               tag=MPI.ANY_TAG,
                               status=status)
            results += result
            comm.send(current_task,
                      dest=status.Get_source(),
                      tag=self._WORKTAG)
            next_start = current_task[0] + current_task[1]
            sending_voxels = self.voxel_unit \
                if self.voxel_unit < self.num_voxels - next_start \
                else self.num_voxels - next_start
            current_task = (next_start, sending_voxels)

        for i in range(0, using_size):
            if i == self.master_rank:
                continue
            result = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG)
            results += result

        for i in range(0, size):
            if i == self.master_rank:
                continue
            comm.send(None,
                      dest=i,
                      tag=self._TERMINATETAG)

        return results

    def _worker(self, clf):
        """Worker node's operation.

        Receiving tasks from the master to process and sending the result back

        Parameters
        ----------
        clf: classification function
            the classifier to be used in cross validation

        Returns
        -------
        None
        """
        logger.debug(
            'worker %d is running, waiting for tasks from master at rank %d' %
            (MPI.COMM_WORLD.Get_rank(), self.master_rank)
        )
        comm = MPI.COMM_WORLD
        status = MPI.Status()
        while 1:
            task = comm.recv(source=self.master_rank,
                             tag=MPI.ANY_TAG,
                             status=status)
            if status.Get_tag():
                break
            comm.send(self._voxel_scoring(task, clf),
                      dest=self.master_rank)

    def _correlation_computation(self, task):
        """Use BLAS API to do correlation computation (matrix multiplication).

        Parameters
        ----------
        task: tuple (start_voxel_id, num_processed_voxels)
            depicting the voxels assigned to compute

        Returns
        -------
        corr: 3D array in shape [num_processed_voxels, num_epochs, num_voxels]
            the correlation values of all subjects in all epochs
            for the assigned values, in row-major
            corr[i, e, s + j] = corr[j, e, s + i]
        """
        time1 = time.time()
        s = task[0]
        nEpochs = len(self.raw_data)
        logger.debug(
            'start to compute the correlation: #epochs: %d, '
            '#processed voxels: %d, #total voxels to compute against: %d' %
            (nEpochs, task[1], self.num_voxels2)
        )
        corr = np.zeros((task[1], nEpochs, self.num_voxels2),
                        np.float32, order='C')
        count = 0
        for i in range(len(self.raw_data)):
            mat = self.raw_data[i]
            mat2 = self.raw_data2[i] if self.raw_data2 is not None else mat
            no_trans = 'N'
            trans = 'T'
            blas.compute_self_corr_for_voxel_sel(no_trans, trans,
                                                 self.num_voxels2, task[1],
                                                 mat.shape[0], 1.0,
                                                 mat2, self.num_voxels2,
                                                 s, mat, self.num_voxels,
                                                 0.0, corr,
                                                 self.num_voxels2 * nEpochs,
                                                 count)
            count += 1
        time2 = time.time()
        logger.debug(
            'correlation computation for %d voxels, takes %.2f s' %
            (task[1], (time2 - time1))
        )
        return corr

    def _correlation_normalization(self, corr):
        """Do within-subject normalization.

        This method uses scipy.zscore to normalize the data,
        but is much slower than its C++ counterpart.
        It is doing in-place z-score.

        Parameters
        ----------
        corr: 3D array in shape [num_processed_voxels, num_epochs, num_voxels]
            the correlation values of all subjects in all epochs
            for the assigned values, in row-major

        Returns
        -------
        corr: 3D array in shape [num_processed_voxels, num_epochs, num_voxels]
            the normalized correlation values of all subjects in all epochs
            for the assigned values, in row-major
        """
        time1 = time.time()
        (sv, e, av) = corr.shape
        for i in range(sv):
            start = 0
            while start < e:
                cur_val = corr[i, start: start + self.epochs_per_subj, :]
                cur_val = .5 * np.log((cur_val + 1) / (1 - cur_val))
                corr[i, start: start + self.epochs_per_subj, :] = \
                    zscore(cur_val, axis=0, ddof=0)
                start += self.epochs_per_subj
        # if zscore fails (standard deviation is zero),
        # set all values to be zero
        corr = np.nan_to_num(corr)
        time2 = time.time()
        logger.debug(
            'within-subject normalization for %d voxels '
            'using numpy zscore function, takes %.2f s' %
            (sv, (time2 - time1))
        )
        return corr

    def _prepare_for_cross_validation(self, corr, clf):
        """Prepare data for voxelwise cross validation.

        If the classifier is sklearn.svm.SVC with precomputed kernel,
        the kernel matrix of each voxel is computed, otherwise do nothing.

        Parameters
        ----------
        corr: 3D array in shape [num_processed_voxels, num_epochs, num_voxels]
            the normalized correlation values of all subjects in all epochs
            for the assigned values, in row-major
        clf: classification function
            the classifier to be used in cross validation

        Returns
        -------
        data: 3D numpy array
            If using sklearn.svm.SVC with precomputed kernel,
            it is in shape [num_processed_voxels, num_epochs, num_epochs];
            otherwise it is the input argument corr,
            in shape [num_processed_voxels, num_epochs, num_voxels]
        """
        time1 = time.time()
        (num_processed_voxels, num_epochs, _) = corr.shape
        if isinstance(clf, sklearn.svm.SVC) and clf.kernel == 'precomputed':
            # kernel matrices should be computed
            kernel_matrices = np.zeros((num_processed_voxels, num_epochs,
                                        num_epochs),
                                       np.float32, order='C')
            for i in range(num_processed_voxels):
                blas.compute_kernel_matrix('L', 'T',
                                           num_epochs, self.num_voxels2,
                                           1.0, corr,
                                           i, self.num_voxels2,
                                           0.0, kernel_matrices[i, :, :],
                                           num_epochs)
                # shrink the values for getting more stable alpha values
                # in SVM training iteration
                num_digits = len(str(int(kernel_matrices[i, 0, 0])))
                if num_digits > 2:
                    proportion = 10**(2-num_digits)
                    kernel_matrices[i, :, :] *= proportion
            data = kernel_matrices
        else:
            data = corr
        time2 = time.time()
        logger.debug(
            'cross validation data preparation takes %.2f s' %
            (time2 - time1)
        )
        return data

    def _do_cross_validation(self, clf, data, task):
        """Run voxelwise cross validation based on correlation vectors.

        clf: classification function
            the classifier to be used in cross validation
        data: 3D numpy array
            If using sklearn.svm.SVC with precomputed kernel,
            it is in shape [num_processed_voxels, num_epochs, num_epochs];
            otherwise it is the input argument corr,
            in shape [num_processed_voxels, num_epochs, num_voxels]
        task: tuple (start_voxel_id, num_processed_voxels)
            depicting the voxels assigned to compute

        Returns
        -------
        results: list of tuple (voxel_id, accuracy)
            the accuracy numbers of all voxels, in accuracy descending order
            the length of array equals the number of assigned voxels
        """
        time1 = time.time()

        if isinstance(clf, sklearn.svm.SVC) and clf.kernel == 'precomputed'\
                and self.use_multiprocessing:
            inlist = [(clf, i + task[0], self.num_folds, data[i, :, :],
                       self.labels) for i in range(task[1])]

            with multiprocessing.Pool(self.process_num) as pool:
                results = list(pool.starmap(_cross_validation_for_one_voxel,
                                            inlist))
        else:
            results = []
            for i in range(task[1]):
                result = _cross_validation_for_one_voxel(clf, i + task[0],
                                                         self.num_folds,
                                                         data[i, :, :],
                                                         self.labels)
                results.append(result)
        time2 = time.time()
        logger.debug(
            'cross validation for %d voxels, takes %.2f s' %
            (task[1], (time2 - time1))
        )
        return results

    def _voxel_scoring(self, task, clf):
        """The voxel selection process done in the worker node.

        Take the task in,
        do analysis on voxels specified by the task (voxel id, num_voxels)
        It is a three-stage pipeline consisting of:
        1. correlation computation
        2. within-subject normalization
        3. voxelwise cross validation

        Parameters
        ----------
        task: tuple (start_voxel_id, num_processed_voxels),
            depicting the voxels assigned to compute
        clf: classification function
            the classifier to be used in cross validation

        Returns
        -------
        results: list of tuple (voxel_id, accuracy)
            the accuracy numbers of all voxels, in accuracy descending order
            the length of array equals the number of assigned voxels
        """
        time1 = time.time()
        # correlation computation
        corr = self._correlation_computation(task)
        # normalization
        # corr = self._correlation_normalization(corr)
        time3 = time.time()
        fcma_extension.normalization(corr, self.epochs_per_subj)
        time4 = time.time()
        logger.debug(
            'within-subject normalization for %d voxels '
            'using C++, takes %.2f s' %
            (task[1], (time4 - time3))
        )

        # cross validation
        data = self._prepare_for_cross_validation(corr, clf)
        if isinstance(clf, sklearn.svm.SVC) and clf.kernel == 'precomputed':
            # to save memory so that the process can be forked
            del corr
        results = self._do_cross_validation(clf, data, task)
        time2 = time.time()
        logger.info(
            'in rank %d, task %d takes %.2f s' %
            (MPI.COMM_WORLD.Get_rank(),
             (int(task[0] / self.voxel_unit)), (time2 - time1))
        )
        return results



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""fMRI Simulator test script

Test script for generating a run of a participant's data.

 Authors: Cameron Ellis (Princeton) 2016
"""
import numpy as np
import math
from brainiak.utils import fmrisim as sim
import pytest
from itertools import product


def test_generate_signal():

    # Inputs for generate_signal
    dimensions = np.array([10, 10, 10])  # What is the size of the brain
    feature_size = [3]
    feature_type = ['cube']
    feature_coordinates = np.array([[5, 5, 5]])
    signal_magnitude = [30]

    # Generate a volume representing the location and quality of the signal
    volume = sim.generate_signal(dimensions=dimensions,
                                 feature_coordinates=feature_coordinates,
                                 feature_type=feature_type,
                                 feature_size=feature_size,
                                 signal_magnitude=signal_magnitude,
                                 )

    assert np.all(volume.shape == dimensions), "Check signal shape"
    assert np.max(volume) == signal_magnitude, "Check signal magnitude"
    assert np.sum(volume > 0) == math.pow(feature_size[0], 3), (
        "Check feature size")
    assert volume[5, 5, 5] == signal_magnitude, "Check signal location"
    assert volume[5, 5, 1] == 0, "Check noise location"

    feature_coordinates = np.array(
        [[5, 5, 5], [3, 3, 3], [7, 7, 7]])

    # Check feature size is correct
    volume = sim.generate_signal(dimensions=dimensions,
                                 feature_coordinates=feature_coordinates,
                                 feature_type=['loop', 'cavity', 'sphere'],
                                 feature_size=[3],
                                 signal_magnitude=signal_magnitude)
    assert volume[5, 5, 5] == 0, "Loop is empty"
    assert volume[3, 3, 3] == 0, "Cavity is empty"
    assert volume[7, 7, 7] != 0, "Sphere is not empty"

    # Check feature size manipulation
    volume = sim.generate_signal(dimensions=dimensions,
                                 feature_coordinates=feature_coordinates,
                                 feature_type=['loop', 'cavity', 'sphere'],
                                 feature_size=[1],
                                 signal_magnitude=signal_magnitude)
    assert volume[5, 6, 6] == 0, "Loop is too big"
    assert volume[3, 5, 5] == 0, "Cavity is too big"
    assert volume[7, 9, 9] == 0, "Sphere is too big"

    # Check that out of bounds feature coordinates are corrected
    feature_coordinates = np.array([0, 2, dimensions[2]])
    x, y, z = sim._insert_idxs(feature_coordinates, feature_size[0],
                               dimensions)
    assert x[1] - x[0] == 2, "x min not corrected"
    assert y[1] - y[0] == 3, "y was corrected when it shouldn't be"
    assert z[1] - z[0] == 1, "z max not corrected"

    # Check that signal patterns are created
    feature_coordinates = np.array([[5, 5, 5]])
    volume = sim.generate_signal(dimensions=dimensions,
                                 feature_coordinates=feature_coordinates,
                                 feature_type=feature_type,
                                 feature_size=feature_size,
                                 signal_magnitude=signal_magnitude,
                                 signal_constant=0,
                                 )
    assert volume[4:7, 4:7, 4:7].std() > 0, "Signal is constant"


def test_generate_stimfunction():

    # Inputs for generate_stimfunction
    onsets = [10, 30, 50, 70, 90]
    event_durations = [6]
    tr_duration = 2
    duration = 100

    # Create the time course for the signal to be generated
    stimfunction = sim.generate_stimfunction(onsets=onsets,
                                             event_durations=event_durations,
                                             total_time=duration,
                                             )

    assert stimfunction.shape[0] == duration * 100, "stimfunc incorrect length"
    eventNumber = np.sum(event_durations * len(onsets)) * 100
    assert np.sum(stimfunction) == eventNumber, "Event number"

    # Create the signal function
    signal_function = sim.convolve_hrf(stimfunction=stimfunction,
                                       tr_duration=tr_duration,
                                       )

    stim_dur = stimfunction.shape[0] / (tr_duration * 100)
    assert signal_function.shape[0] == stim_dur, "The length did not change"

    # Test
    onsets = [0]
    tr_duration = 1
    event_durations = [1]
    stimfunction = sim.generate_stimfunction(onsets=onsets,
                                             event_durations=event_durations,
                                             total_time=duration,
                                             )

    signal_function = sim.convolve_hrf(stimfunction=stimfunction,
                                       tr_duration=tr_duration,
                                       )

    max_response = np.where(signal_function != 0)[0].max()
    assert 25 < max_response <= 30, "HRF has the incorrect length"
    assert np.sum(signal_function < 0) > 0, "No values below zero"

    # Export a stimfunction
    sim.export_3_column(stimfunction,
                        'temp.txt',
                        )

    # Load in the stimfunction
    stimfunc_new = sim.generate_stimfunction(onsets=None,
                                             event_durations=None,
                                             total_time=duration,
                                             timing_file='temp.txt',
                                             )

    assert np.all(stimfunc_new == stimfunction), "Export/import failed"

    # Break the timing precision of the generation
    stimfunc_new = sim.generate_stimfunction(onsets=None,
                                             event_durations=None,
                                             total_time=duration,
                                             timing_file='temp.txt',
                                             temporal_resolution=0.5,
                                             )

    assert stimfunc_new.sum() == 0, "Temporal resolution not working right"

    # Set the duration to be too short so you should get an error
    onsets = [10, 30, 50, 70, 90]
    event_durations = [5]
    with pytest.raises(ValueError):
        sim.generate_stimfunction(onsets=onsets,
                                  event_durations=event_durations,
                                  total_time=89,
                                  )

    # Clip the event offset
    stimfunc_new = sim.generate_stimfunction(onsets=onsets,
                                             event_durations=event_durations,
                                             total_time=95,
                                             )
    assert stimfunc_new[-1] == 1, 'Event offset was not clipped'

    # Test exporting a group of participants to an epoch file
    cond_a = sim.generate_stimfunction(onsets=onsets,
                                       event_durations=event_durations,
                                       total_time=110,
                                       )

    cond_b = sim.generate_stimfunction(onsets=[x + 5 for x in onsets],
                                       event_durations=event_durations,
                                       total_time=110,
                                       )

    stimfunction_group = [np.hstack((cond_a, cond_b))] * 2
    sim.export_epoch_file(stimfunction_group,
                          'temp.txt',
                          tr_duration,
                          )

    # Check that convolve throws a warning when the shape is wrong
    sim.convolve_hrf(stimfunction=np.hstack((cond_a, cond_b)).T,
                     tr_duration=tr_duration,
                     temporal_resolution=1,
                     )


def test_apply_signal():

    dimensions = np.array([10, 10, 10])  # What is the size of the brain
    feature_size = [2]
    feature_type = ['cube']
    feature_coordinates = np.array(
        [[5, 5, 5]])
    signal_magnitude = [30]

    # Generate a volume representing the location and quality of the signal
    volume = sim.generate_signal(dimensions=dimensions,
                                 feature_coordinates=feature_coordinates,
                                 feature_type=feature_type,
                                 feature_size=feature_size,
                                 signal_magnitude=signal_magnitude,
                                 )

    # Inputs for generate_stimfunction
    onsets = [10, 30, 50, 70, 90]
    event_durations = [6]
    tr_duration = 2
    duration = 100

    # Create the time course for the signal to be generated
    stimfunction = sim.generate_stimfunction(onsets=onsets,
                                             event_durations=event_durations,
                                             total_time=duration,
                                             )

    signal_function = sim.convolve_hrf(stimfunction=stimfunction,
                                       tr_duration=tr_duration,
                                       )

    # Check that you can compute signal change appropriately
    # Preset a bunch of things
    stimfunction_tr = stimfunction[::int(tr_duration * 100)]
    mask, template = sim.mask_brain(dimensions, mask_self=False)
    noise_dict = sim._noise_dict_update({})
    noise = sim.generate_noise(dimensions=dimensions,
                               stimfunction_tr=stimfunction_tr,
                               tr_duration=tr_duration,
                               template=template,
                               mask=mask,
                               noise_dict=noise_dict,
                               iterations=[0, 0]
                               )
    coords = feature_coordinates[0]
    noise_function_a = noise[coords[0], coords[1], coords[2], :]
    noise_function_a = noise_function_a.reshape(duration // tr_duration, 1)

    noise_function_b = noise[coords[0] + 1, coords[1], coords[2], :]
    noise_function_b = noise_function_b.reshape(duration // tr_duration, 1)

    # Check that the noise_function and signal_function must be the same size
    method = 'PSC'
    with pytest.raises(ValueError):
        sim.compute_signal_change(signal_function,
                                  noise_function_a.T,
                                  noise_dict,
                                  [0.5],
                                  method,
                                  )

    # Create the calibrated signal with PSC
    sig_a = sim.compute_signal_change(signal_function,
                                      noise_function_a,
                                      noise_dict,
                                      [0.5],
                                      method,
                                      )
    sig_b = sim.compute_signal_change(signal_function,
                                      noise_function_a,
                                      noise_dict,
                                      [1.0],
                                      method,
                                      )

    assert sig_b.max() / sig_a.max() == 2, 'PSC modulation failed'

    # Create the calibrated signal with SFNR
    method = 'SFNR'
    sig_a = sim.compute_signal_change(signal_function,
                                      noise_function_a,
                                      noise_dict,
                                      [0.5],
                                      method,
                                      )
    scaled_a = sig_a / (noise_function_a.mean() / noise_dict['sfnr'])
    sig_b = sim.compute_signal_change(signal_function,
                                      noise_function_b,
                                      noise_dict,
                                      [1.0],
                                      method,
                                      )
    scaled_b = sig_b / (noise_function_b.mean() / noise_dict['sfnr'])

    assert scaled_b.max() / scaled_a.max() == 2, 'SFNR modulation failed'

    # Create the calibrated signal with CNR_Amp/Noise-SD
    method = 'CNR_Amp/Noise-SD'
    sig_a = sim.compute_signal_change(signal_function,
                                      noise_function_a,
                                      noise_dict,
                                      [0.5],
                                      method,
                                      )
    scaled_a = sig_a / noise_function_a.std()
    sig_b = sim.compute_signal_change(signal_function,
                                      noise_function_b,
                                      noise_dict,
                                      [1.0],
                                      method,
                                      )
    scaled_b = sig_b / noise_function_b.std()

    assert scaled_b.max() / scaled_a.max() == 2, 'CNR_Amp modulation failed'

    # Create the calibrated signal with CNR_Amp/Noise-Var_dB
    method = 'CNR_Amp2/Noise-Var_dB'
    sig_a = sim.compute_signal_change(signal_function,
                                      noise_function_a,
                                      noise_dict,
                                      [0.5],
                                      method,
                                      )
    scaled_a = np.log(sig_a.max() / noise_function_a.std())
    sig_b = sim.compute_signal_change(signal_function,
                                      noise_function_b,
                                      noise_dict,
                                      [1.0],
                                      method,
                                      )
    scaled_b = np.log(sig_b.max() / noise_function_b.std())

    assert np.round(scaled_b / scaled_a) == 2, 'CNR_Amp dB modulation failed'

    # Create the calibrated signal with CNR_Signal-SD/Noise-SD
    method = 'CNR_Signal-SD/Noise-SD'
    sig_a = sim.compute_signal_change(signal_function,
                                      noise_function_a,
                                      noise_dict,
                                      [0.5],
                                      method,
                                      )
    scaled_a = sig_a.std() / noise_function_a.std()
    sig_b = sim.compute_signal_change(signal_function,
                                      noise_function_a,
                                      noise_dict,
                                      [1.0],
                                      method,
                                      )
    scaled_b = sig_b.std() / noise_function_a.std()

    assert (scaled_b / scaled_a) == 2, 'CNR signal modulation failed'

    # Create the calibrated signal with CNR_Amp/Noise-Var_dB
    method = 'CNR_Signal-Var/Noise-Var_dB'
    sig_a = sim.compute_signal_change(signal_function,
                                      noise_function_a,
                                      noise_dict,
                                      [0.5],
                                      method,
                                      )

    scaled_a = np.log(sig_a.std() / noise_function_a.std())
    sig_b = sim.compute_signal_change(signal_function,
                                      noise_function_b,
                                      noise_dict,
                                      [1.0],
                                      method,
                                      )
    scaled_b = np.log(sig_b.std() / noise_function_b.std())

    assert np.round(scaled_b / scaled_a) == 2, 'CNR signal dB modulation ' \
                                               'failed'

    # Convolve the HRF with the stimulus sequence
    signal = sim.apply_signal(signal_function=signal_function,
                              volume_signal=volume,
                              )

    assert signal.shape == (dimensions[0], dimensions[1], dimensions[2],
                            duration / tr_duration), "The output is the " \
                                                     "wrong size"

    signal = sim.apply_signal(signal_function=stimfunction,
                              volume_signal=volume,
                              )

    assert np.any(signal == signal_magnitude), "The stimfunction is not binary"

    # Check that there is an error if the number of signal voxels doesn't
    # match the number of non zero brain voxels
    with pytest.raises(IndexError):
        sig_vox = (volume > 0).sum()
        vox_pattern = np.tile(stimfunction, (1, sig_vox - 1))
        sim.apply_signal(signal_function=vox_pattern,
                         volume_signal=volume,
                         )


def test_generate_noise():

    dimensions = np.array([10, 10, 10])  # What is the size of the brain
    feature_size = [2]
    feature_type = ['cube']
    feature_coordinates = np.array(
        [[5, 5, 5]])
    signal_magnitude = [1]

    # Generate a volume representing the location and quality of the signal
    volume = sim.generate_signal(dimensions=dimensions,
                                 feature_coordinates=feature_coordinates,
                                 feature_type=feature_type,
                                 feature_size=feature_size,
                                 signal_magnitude=signal_magnitude,
                                 )

    # Inputs for generate_stimfunction
    onsets = [10, 30, 50, 70, 90]
    event_durations = [6]
    tr_duration = 2
    duration = 200

    # Create the time course for the signal to be generated
    stimfunction = sim.generate_stimfunction(onsets=onsets,
                                             event_durations=event_durations,
                                             total_time=duration,
                                             )

    signal_function = sim.convolve_hrf(stimfunction=stimfunction,
                                       tr_duration=tr_duration,
                                       )

    # Convolve the HRF with the stimulus sequence
    signal = sim.apply_signal(signal_function=signal_function,
                              volume_signal=volume,
                              )

    # Generate the mask of the signal
    mask, template = sim.mask_brain(signal,
                                    mask_self=None)

    assert min(mask[mask > 0]) > 0.1, "Mask thresholding did not work"
    assert len(np.unique(template) > 2), "Template creation did not work"

    stimfunction_tr = stimfunction[::int(tr_duration * 100)]

    # Create the noise volumes (using the default parameters)
    noise = sim.generate_noise(dimensions=dimensions,
                               stimfunction_tr=stimfunction_tr,
                               tr_duration=tr_duration,
                               template=template,
                               mask=mask,
                               iterations=[1, 0],
                               )

    assert signal.shape == noise.shape, "The dimensions of signal and noise " \
                                        "the same"

    noise_high = sim.generate_noise(dimensions=dimensions,
                                    stimfunction_tr=stimfunction_tr,
                                    tr_duration=tr_duration,
                                    template=template,
                                    mask=mask,
                                    noise_dict={'sfnr': 50, 'snr': 25},
                                    iterations=[1, 0],
                                    )

    noise_low = sim.generate_noise(dimensions=dimensions,
                                   stimfunction_tr=stimfunction_tr,
                                   tr_duration=tr_duration,
                                   template=template,
                                   mask=mask,
                                   noise_dict={'sfnr': 100, 'snr': 25},
                                   iterations=[1, 0],
                                   )

    system_high = np.std(noise_high[mask > 0], 1).mean()
    system_low = np.std(noise_low[mask > 0], 1).mean()

    assert system_low < system_high, "SFNR noise could not be manipulated"

    # Check that you check for the appropriate template values
    with pytest.raises(ValueError):
        sim.generate_noise(dimensions=dimensions,
                           stimfunction_tr=stimfunction_tr,
                           tr_duration=tr_duration,
                           template=template * 2,
                           mask=mask,
                           noise_dict={},
                           )

    # Check that iterations does what it should
    sim.generate_noise(dimensions=dimensions,
                       stimfunction_tr=stimfunction_tr,
                       tr_duration=tr_duration,
                       template=template,
                       mask=mask,
                       noise_dict={},
                       iterations=[0, 0],
                       )

    sim.generate_noise(dimensions=dimensions,
                       stimfunction_tr=stimfunction_tr,
                       tr_duration=tr_duration,
                       template=template,
                       mask=mask,
                       noise_dict={},
                       iterations=None,
                       )

    # Test drift noise
    trs = 1000
    period = 100
    drift = sim._generate_noise_temporal_drift(trs,
                                               tr_duration,
                                               'sine',
                                               period,
                                               )

    # Check that the max frequency is the appropriate frequency
    power = abs(np.fft.fft(drift))[1:trs // 2]
    freq = np.linspace(1, trs // 2 - 1, trs // 2 - 1) / trs
    period_freq = np.where(freq == 1 / (period // tr_duration))
    max_freq = np.argmax(power)

    assert period_freq == max_freq, 'Max frequency is not where it should be'

    # Do the same but now with cosine basis functions, answer should be close
    drift = sim._generate_noise_temporal_drift(trs,
                                               tr_duration,
                                               'discrete_cos',
                                               period,
                                               )

    # Check that the appropriate frequency is peaky (may not be the max)
    power = abs(np.fft.fft(drift))[1:trs // 2]
    freq = np.linspace(1, trs // 2 - 1, trs // 2 - 1) / trs
    period_freq = np.where(freq == 1 / (period // tr_duration))[0][0]

    assert power[period_freq] > power[period_freq + 1], 'Power is low'
    assert power[period_freq] > power[period_freq - 1], 'Power is low'

    # Check it runs fine
    drift = sim._generate_noise_temporal_drift(50,
                                               tr_duration,
                                               'discrete_cos',
                                               period,
                                               )

    # Check it runs fine
    drift = sim._generate_noise_temporal_drift(300,
                                               tr_duration,
                                               'cos_power_drop',
                                               period,
                                               )

    # Check that when the TR is greater than the period it errors
    with pytest.raises(ValueError):
        sim._generate_noise_temporal_drift(30, 10, 'cos_power_drop', 5)

    # Test physiological noise (using unrealistic parameters so that it's easy)
    timepoints = list(np.linspace(0, (trs - 1) * tr_duration, trs))
    resp_freq = 0.2
    heart_freq = 1.17
    phys = sim._generate_noise_temporal_phys(timepoints,
                                             resp_freq,
                                             heart_freq,
                                             )

    # Check that the max frequency is the appropriate frequency
    power = abs(np.fft.fft(phys))[1:trs // 2]
    freq = np.linspace(1, trs // 2 - 1, trs // 2 - 1) / (trs * tr_duration)
    peaks = (power > (power.mean() + power.std()))  # Where are the peaks
    peak_freqs = freq[peaks]

    assert np.any(resp_freq == peak_freqs), 'Resp frequency not found'
    assert len(peak_freqs) == 2, 'Two peaks not found'

    # Test task noise
    sim._generate_noise_temporal_task(stimfunction_tr,
                                      motion_noise='gaussian',
                                      )
    sim._generate_noise_temporal_task(stimfunction_tr,
                                      motion_noise='rician',
                                      )

    # Test ARMA noise
    with pytest.raises(ValueError):
        noise_dict = {'fwhm': 4, 'auto_reg_rho': [1], 'ma_rho': [1, 1]}
        sim._generate_noise_temporal_autoregression(stimfunction_tr,
                                                    noise_dict,
                                                    dimensions,
                                                    mask,
                                                    )

    # Generate spatial noise
    vol = sim._generate_noise_spatial(np.array([10, 10, 10, trs]))
    assert len(vol.shape) == 3, 'Volume was not reshaped to ignore TRs'

    # Switch some of the noise types on
    noise_dict = dict(physiological_sigma=1, drift_sigma=1, task_sigma=1,
                      auto_reg_sigma=0)
    sim.generate_noise(dimensions=dimensions,
                       stimfunction_tr=stimfunction_tr,
                       tr_duration=tr_duration,
                       template=template,
                       mask=mask,
                       noise_dict=noise_dict,
                       iterations=[0, 0],
                       )


def test_generate_noise_spatial():

    # Set up the inputs
    dimensions = np.array([10, 5, 10])
    mask = np.ones(dimensions)
    vol = sim._generate_noise_spatial(dimensions, mask)

    # Run the analysis from _calc_FHWM but for th elast step of aggregating
    # across dimensions
    v_count = 0
    v_sum = 0
    v_sq = 0

    d_sum = [0.0, 0.0, 0.0]
    d_sq = [0.0, 0.0, 0.0]
    d_count = [0, 0, 0]

    # Pull out all the voxel coordinates
    coordinates = list(product(range(dimensions[0]),
                               range(dimensions[1]),
                               range(dimensions[2])))

    # Find the sum of squared error for the non-masked voxels in the brain
    for i in list(range(len(coordinates))):

        # Pull out this coordinate
        x, y, z = coordinates[i]

        # Is this within the mask?
        if mask[x, y, z] > 0:

            # Find the the volume sum and squared values
            v_count += 1
            v_sum += vol[x, y, z]
            v_sq += vol[x, y, z] ** 2

    # Get the volume variance
    v_var = (v_sq - ((v_sum ** 2) / v_count)) / (v_count - 1)

    for i in list(range(len(coordinates))):

        # Pull out this coordinate
        x, y, z = coordinates[i]

        # Is this within the mask?
        if mask[x, y, z] > 0:
            # For each xyz dimension calculate the squared
            # difference of this voxel and the next

            in_range = (x < dimensions[0] - 1)
            in_mask = in_range and (mask[x + 1, y, z] > 0)
            included = in_mask and (~np.isnan(vol[x + 1, y, z]))
            if included:
                d_sum[0] += vol[x, y, z] - vol[x + 1, y, z]
                d_sq[0] += (vol[x, y, z] - vol[x + 1, y, z]) ** 2
                d_count[0] += 1

            in_range = (y < dimensions[1] - 1)
            in_mask = in_range and (mask[x, y + 1, z] > 0)
            included = in_mask and (~np.isnan(vol[x, y + 1, z]))
            if included:
                d_sum[1] += vol[x, y, z] - vol[x, y + 1, z]
                d_sq[1] += (vol[x, y, z] - vol[x, y + 1, z]) ** 2
                d_count[1] += 1

            in_range = (z < dimensions[2] - 1)
            in_mask = in_range and (mask[x, y, z + 1] > 0)
            included = in_mask and (~np.isnan(vol[x, y, z + 1]))
            if included:
                d_sum[2] += vol[x, y, z] - vol[x, y, z + 1]
                d_sq[2] += (vol[x, y, z] - vol[x, y, z + 1]) ** 2
                d_count[2] += 1

    # Find the variance
    d_var = np.divide((d_sq - np.divide(np.power(d_sum, 2),
                                        d_count)), (np.add(d_count, -1)))

    o_var = np.divide(-1, (4 * np.log(1 - (0.5 * d_var / v_var))))
    fwhm3 = np.sqrt(o_var) * 2 * np.sqrt(2 * np.log(2))

    # Calculate the proportion of std relative to the mean
    std_proportion = np.nanstd(fwhm3) / np.nanmean(fwhm3)
    assert std_proportion < 0.25, 'Variance is inconsistent across dim'


def test_mask_brain():

    # Inputs for generate_signal
    dimensions = np.array([10, 10, 10])  # What is the size of the brain
    feature_size = [2]
    feature_type = ['cube']
    feature_coordinates = np.array(
        [[4, 4, 4]])
    signal_magnitude = [30]

    # Generate a volume representing the location and quality of the signal
    volume = sim.generate_signal(dimensions=dimensions,
                                 feature_coordinates=feature_coordinates,
                                 feature_type=feature_type,
                                 feature_size=feature_size,
                                 signal_magnitude=signal_magnitude,
                                 )

    # Mask the volume to be the same shape as a brain
    mask, _ = sim.mask_brain(dimensions, mask_self=None,)
    brain = volume * mask

    assert np.sum(brain != 0) == np.sum(volume != 0), "Masking did not work"
    assert brain[0, 0, 0] == 0, "Masking did not work"
    assert brain[4, 4, 4] != 0, "Masking did not work"

    feature_coordinates = np.array(
        [[1, 1, 1]])

    volume = sim.generate_signal(dimensions=dimensions,
                                 feature_coordinates=feature_coordinates,
                                 feature_type=feature_type,
                                 feature_size=feature_size,
                                 signal_magnitude=signal_magnitude,
                                 )

    # Mask the volume to be the same shape as a brain
    mask, _ = sim.mask_brain(dimensions, mask_self=None, )
    brain = volume * mask

    assert np.sum(brain != 0) < np.sum(volume != 0), "Masking did not work"

    # Test that you can load the default
    dimensions = np.array([100, 100, 100])
    mask, template = sim.mask_brain(dimensions, mask_self=False)

    assert mask[20, 80, 50] == 0, 'Masking didn''t work'
    assert mask[25, 80, 50] == 1, 'Masking didn''t work'
    assert int(template[25, 80, 50] * 100) == 57, 'Template not correct'

    # Check that you can mask self
    mask_self, template_self = sim.mask_brain(template, mask_self=True)

    assert (template_self - template).sum() < 1e2, 'Mask self error'
    assert (mask_self - mask).sum() == 0, 'Mask self error'


def test_calc_noise():

    # Inputs for functions
    onsets = [10, 30, 50, 70, 90]
    event_durations = [6]
    tr_duration = 2
    duration = 200
    temporal_res = 100
    tr_number = int(np.floor(duration / tr_duration))
    dimensions_tr = np.array([10, 10, 10, tr_number])

    # Preset the noise dict
    nd_orig = sim._noise_dict_update({})

    # Create the time course for the signal to be generated
    stimfunction = sim.generate_stimfunction(onsets=onsets,
                                             event_durations=event_durations,
                                             total_time=duration,
                                             temporal_resolution=temporal_res,
                                             )

    # Mask the volume to be the same shape as a brain
    mask, template = sim.mask_brain(dimensions_tr, mask_self=None)
    stimfunction_tr = stimfunction[::int(tr_duration * temporal_res)]

    nd_orig['matched'] = 0
    noise = sim.generate_noise(dimensions=dimensions_tr[0:3],
                               stimfunction_tr=stimfunction_tr,
                               tr_duration=tr_duration,
                               template=template,
                               mask=mask,
                               noise_dict=nd_orig,
                               )

    # Check the spatial noise match
    nd_orig['matched'] = 1
    noise_matched = sim.generate_noise(dimensions=dimensions_tr[0:3],
                                       stimfunction_tr=stimfunction_tr,
                                       tr_duration=tr_duration,
                                       template=template,
                                       mask=mask,
                                       noise_dict=nd_orig,
                                       iterations=[50, 0]
                                       )

    # Calculate the noise parameters from this newly generated volume
    nd_new = sim.calc_noise(noise, mask, template)
    nd_matched = sim.calc_noise(noise_matched, mask, template)

    # Check the values are reasonable"
    assert nd_new['snr'] > 0, 'snr out of range'
    assert nd_new['sfnr'] > 0, 'sfnr out of range'
    assert nd_new['auto_reg_rho'][0] > 0, 'ar out of range'

    # Check that the dilation increases SNR
    no_dilation_snr = sim._calc_snr(noise_matched,
                                    mask,
                                    dilation=0,
                                    reference_tr=tr_duration,
                                    )

    assert nd_new['snr'] > no_dilation_snr, "Dilation did not increase SNR"

    # Check that template size is in bounds
    with pytest.raises(ValueError):
        sim.calc_noise(noise, mask, template * 2)

    # Check that Mask is set is checked
    with pytest.raises(ValueError):
        sim.calc_noise(noise, None, template)

    # Check that it can deal with missing noise parameters
    temp_nd = sim.calc_noise(noise, mask, template, noise_dict={})
    assert temp_nd['voxel_size'][0] == 1, 'Default voxel size not set'

    temp_nd = sim.calc_noise(noise, mask, template, noise_dict=None)
    assert temp_nd['voxel_size'][0] == 1, 'Default voxel size not set'

    # Check that the fitting worked
    snr_diff = abs(nd_orig['snr'] - nd_new['snr'])
    snr_diff_match = abs(nd_orig['snr'] - nd_matched['snr'])
    assert snr_diff > snr_diff_match, 'snr fit incorrectly'

    # Test that you can generate rician and exponential noise
    sim._generate_noise_system(dimensions_tr,
                               1,
                               1,
                               spatial_noise_type='exponential',
                               temporal_noise_type='rician',
                               )

    # Check the temporal noise match
    nd_orig['matched'] = 1
    noise_matched = sim.generate_noise(dimensions=dimensions_tr[0:3],
                                       stimfunction_tr=stimfunction_tr,
                                       tr_duration=tr_duration,
                                       template=template,
                                       mask=mask,
                                       noise_dict=nd_orig,
                                       iterations=[0, 50]
                                       )

    nd_matched = sim.calc_noise(noise_matched, mask, template)

    sfnr_diff = abs(nd_orig['sfnr'] - nd_new['sfnr'])
    sfnr_diff_match = abs(nd_orig['sfnr'] - nd_matched['sfnr'])
    assert sfnr_diff > sfnr_diff_match, 'sfnr fit incorrectly'

    ar1_diff = abs(nd_orig['auto_reg_rho'][0] - nd_new['auto_reg_rho'][0])
    ar1_diff_match = abs(nd_orig['auto_reg_rho'][0] - nd_matched[
        'auto_reg_rho'][0])
    assert ar1_diff > ar1_diff_match, 'AR1 fit incorrectly'

    # Check that you can calculate ARMA for a single voxel
    vox = noise[5, 5, 5, :]
    arma = sim._calc_ARMA_noise(vox,
                                None,
                                sample_num=2,
                                )
    assert len(arma) == 2, "Two outputs not given by ARMA"


def test_gen_1D_gauss_shape():
    n_vox = 10
    res = 180
    rfs, centers = sim.generate_1d_gaussian_rfs(n_vox, res, (0, res-1))
    assert rfs.shape == (n_vox, res)
    assert centers.size == n_vox

    sim_data = sim.generate_1d_rf_responses(rfs, np.array([0, 10, 20]), res,
                                            (0, res-1))
    assert sim_data.shape == (n_vox, 3)


def test_gen_1d_gauss_range():
    res = 180
    range_values = (-10, res-11)
    rfs, centers = sim.generate_1d_gaussian_rfs(1, res, range_values,
                                                random_tuning=False)
    sim_data = sim.generate_1d_rf_responses(rfs, np.array([-10]), res,
                                            range_values, 0)
    assert sim_data[0, ] > 0
    range_values = (10, res+10)
    rfs, centers = sim.generate_1d_gaussian_rfs(1, res, range_values,
                                                random_tuning=False)
    sim_data = sim.generate_1d_rf_responses(rfs, np.array([10]), res,
                                            range_values, 0)
    assert sim_data[0, ] > 0


def test_gen_1D_gauss_even_spacing():
    n_vox = 9
    res = 180
    rfs, centers = sim.generate_1d_gaussian_rfs(n_vox, res, (0, res-1),
                                                random_tuning=False)
    assert np.all(centers == np.array([0, 19, 39, 59, 79, 99, 119, 139, 159]))



import tensorflow as tf
import tensorflow_probability as tfp
from scipy.stats import norm
from numpy.linalg import cholesky
import numpy as np


def rmn(rowcov, colcov):
    """
    Generate random draws from a zero-mean matrix-normal distribution.

    Parameters
    -----------
    rowcov : np.ndarray
        Row covariance (assumed to be positive definite)
    colcov : np.ndarray
        Column covariance (assumed to be positive definite)
    """

    Z = norm.rvs(size=(rowcov.shape[0], colcov.shape[0]))
    return cholesky(rowcov).dot(Z).dot(cholesky(colcov))


def xx_t(x):
    """
    Outer product
    :math:`xx^T`

    Parameters
    -----------
    x : tf.Tensor

    """
    return tf.matmul(x, x, transpose_b=True)


def x_tx(x):
    """Inner product
    :math:`x^T x`

    Parameters
    -----------
    x : tf.Tensor

    """
    return tf.matmul(x, x, transpose_a=True)


def scaled_I(x, size):
    """Scaled identity matrix
    :math:`x I_{size}`

    Parameters
    ------------
    x: float or coercable to float
        Scale to multiply the identity matrix by
    size: int or otherwise coercable to a size
        Dimension of the scaled identity matrix to return
    """
    return tf.linalg.tensor_diag(tf.ones([size], dtype=tf.float64) * x)


def flatten_cholesky_unique(L):
    """
    Flattens nonzero-elements Cholesky (triangular) factor
    into a vector, and logs diagonal to make parameterization
    unique. Inverse of unflatten_cholesky_unique.
    """
    L_tf = tf.linalg.set_diag(L, tf.math.log(tf.linalg.diag_part(L)))
    L_flat = tfp.math.fill_triangular_inverse(L_tf)
    return L_flat


def unflatten_cholesky_unique(L_flat):
    """
    Converts a vector of elements into a triangular matrix
    (Cholesky factor). Exponentiates diagonal to make
    parameterization unique. Inverse of flatten_cholesky_unique.
    """
    L = tfp.math.fill_triangular(L_flat)
    # exp diag for unique parameterization
    L = tf.linalg.set_diag(L, tf.exp(tf.linalg.diag_part(L)))
    return L


def pack_trainable_vars(trainable_vars):
    """
    Pack trainable vars in a model into a single
    vector that can be passed to scipy.optimize
    """
    return tf.concat([tf.reshape(tv, (-1,)) for tv in trainable_vars], axis=0)


def unpack_trainable_vars(x, trainable_vars):
    """
    Unpack trainable vars from a single vector as
    used/returned by scipy.optimize
    """

    sizes = [tv.shape for tv in trainable_vars]
    idxs = [np.prod(sz) for sz in sizes]
    flatvars = tf.split(x, idxs)
    return [tf.reshape(fv, tv.shape) for fv, tv in zip(flatvars,
                                                       trainable_vars)]


def make_val_and_grad(lossfn, train_vars):
    """
    Makes a function that ouptuts the loss and gradient in a format compatible
    with scipy.optimize.minimize
    """

    def val_and_grad(theta):
        with tf.GradientTape() as tape:
            tape.watch(train_vars)
            unpacked_theta = unpack_trainable_vars(theta, train_vars)
            for var, val in zip(train_vars, unpacked_theta):
                var.assign(val)
            loss = lossfn(theta)
        grad = tape.gradient(loss, train_vars)
        packed_grad = pack_trainable_vars(grad)
        return loss.numpy(), packed_grad.numpy()

    return val_and_grad



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import scipy.io
from scipy.stats import stats
import numpy as np
import nibabel as nib
from nilearn.input_data import NiftiMasker
import os
import logging
import sys
import click

@click.command()
@click.argument('nifti-file', type=click.File('rb'))
@click.argument('out-file', type=click.File('wb'))
@click.option('--mask-file', default=None, type=click.File('rb'), help='The mask file to get ROI')
@click.option('--zscore', is_flag=True, help='to zscore fMRI data')
@click.option('--detrend', is_flag=True, help='to detrend fMRI data')
@click.option('--smoothing-fwmw', default=False, type=float, help='the spatial smooth window size')



def extract_data(nifti_file, mask_file, out_file, zscore, detrend, smoothing_fwmw):
    if mask_file is None:
        #whole brain, get coordinate info from nifti_file itself
        mask = nib.load(nifti_file.name)
    else:
        mask = nib.load(mask_file.name)
    affine = mask.get_affine()
    if mask_file is None:
        mask_data = mask.get_fdata()
        if mask_data.ndim == 4:
            #get mask in 3D
            img_data_type = mask.header.get_data_dtype()
            n_tr = mask_data.shape[3]
            mask_data = mask_data[:,:,:,n_tr//2].astype(bool)
            mask = nib.Nifti1Image(mask_data.astype(img_data_type), affine)
        else:
            mask_data = mask_data.astype(bool)
    else:
        mask_data = mask.get_fdata().astype(bool)

    #get voxel coordinates
    R = np.float64(np.argwhere(mask_data))

    #get scanner RAS coordinates based on voxel coordinates
    if affine is not []:
        R = (np.dot(affine[:3,:3], R.T) + affine[:3,3:4]).T

    #get ROI data, and run preprocessing
    nifti_masker = NiftiMasker(mask_img=mask, standardize=zscore, detrend=detrend, smoothing_fwhm=smoothing_fwmw)
    img = nib.load(nifti_file.name)
    all_images = np.float64(nifti_masker.fit_transform(img))
    data = all_images.T.copy()

    #save data
    subj_data = {'data': data, 'R': R}
    scipy.io.savemat(out_file.name, subj_data)


if __name__ == '__main__':
    extract_data()




"""
A pytest plugin to run tests with mpiexec who are marked with the `mpiexec`
marker.

Taken from: https://github.com/minrk/pytest-mpiexec

Slightly modified to work with BrainIAK's pytest configuration.

"""

import json
import os
import shlex
import subprocess
import sys
from enum import Enum
from functools import partial
from itertools import chain
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from pytest_reportlog.plugin import ReportLogPlugin

MPI_SUBPROCESS_ENV = "TEST_MPI_SUBTEST"
TEST_REPORT_DIR_ENV = "TEST_MPI_REPORT_DIR"

MPI_MARKER_NAME = "mpiexec"

MPIEXEC = "mpiexec"


def pytest_addoption(parser):
    group = parser.getgroup("mpiexec")
    group.addoption(
        "--mpiexec",
        action="store",
        dest="mpiexec",
        default=MPIEXEC,
        help="Executable for running MPI, default=mpiexec",
    )
    group.addoption(
        "--mpiexec-report",
        action="store",
        dest="mpiexec_report",
        choices=[r.value for r in ReportStyle],
        default=ReportStyle.first_failure,
        help="""style of mpi error reporting.

        Since each mpi test represents one test run per rank,
        there are lots of ways to represent a failed parallel run:

        Options:

        - first_failure (default): report only one result per test,
          PASSED or FAILED, where FAILED will be the failure of the first
          rank that failed.
        - all_failures: report failures from all ranks that failed
        - all: report all results, including all passes
        - concise: like first_failure, but try to report all _unique_
          failures (experimental)
        """,
    )


def pytest_configure(config):
    global MPIEXEC
    global REPORT_STYLE
    mpiexec = config.getoption("mpiexec")
    if mpiexec:
        MPIEXEC = mpiexec

    REPORT_STYLE = config.getoption("mpiexec_report")

    config.addinivalue_line("markers",
                            f"{MPI_MARKER_NAME}: Run this text with mpiexec")
    if os.getenv(MPI_SUBPROCESS_ENV):
        from mpi4py import MPI

        rank = MPI.COMM_WORLD.rank
        reportlog_dir = Path(os.getenv(TEST_REPORT_DIR_ENV, ""))
        report_path = reportlog_dir / f"reportlog-{rank}.jsonl"
        config._mpiexec_reporter = reporter = (
            ReportLogPlugin(config, report_path))
        config.pluginmanager.register(reporter)


def pytest_unconfigure(config):
    reporter = getattr(config, "_mpiexec_reporter", None)
    if reporter:
        reporter.close()


def mpi_runtest_protocol(item):
    """The runtest protocol for mpi tests

    Runs the test in an mpiexec subprocess

    instead of the current process
    """
    hook = item.config.hook
    hook.pytest_runtest_logstart(nodeid=item.nodeid, location=item.location)
    call = pytest.CallInfo.from_call(
        partial(mpi_runtest, item), "setup")
    if call.excinfo:
        report = hook.pytest_runtest_makereport(item=item, call=call)
        hook.pytest_runtest_logreport(report=report)
    hook.pytest_runtest_logfinish(nodeid=item.nodeid, location=item.location)


def pytest_runtest_protocol(item, nextitem):
    """Run the MPI protocol for mpi tests

    otherwise, do nothing
    """
    if os.getenv(MPI_SUBPROCESS_ENV):
        return
    mpi_mark = item.get_closest_marker(MPI_MARKER_NAME)
    if not mpi_mark:
        return
    mpi_runtest_protocol(item)
    return True


class ReportStyle(Enum):
    all = "all"
    all_failures = "all_failures"
    first_failure = "first_failure"
    concise = "concise"


def _report_key(report):
    """Determine if a given report has been 'seen' before"""
    # use reprcrash for 'same' error message

    message_key = None
    if report["outcome"] != "passed":
        # for failures, use first line of reprcrash
        # (i.e. the line used)
        longrepr = report["longrepr"]
        if longrepr:
            reprcrash = longrepr["reprcrash"]
            if reprcrash:
                message_key_items = []
                for key, value in sorted(
                        report["longrepr"]["reprcrash"].items()):
                    if key == "message":
                        value = value.splitlines()[0]
                    message_key_items.append((key, value))
                message_key = tuple(message_key_items)

    if not message_key and report["outcome"] != "passed":
        # warn about missing message key?
        # warnings.warn("Expected reprcrash...", RuntimeWarning, stacklevel=2)
        pass
    return (report["when"], report["outcome"], message_key)


def consolidate_reports(nodeid, reports, style=ReportStyle.first_failure):
    """Consolidate a collection of TestReports

    - collapses to single success if all succeed
    - all_failures reports all failures
    - first_failure reports only the first failure
    """
    style = ReportStyle(style)

    all_ranks = {report["_mpi_rank"] for report in reports}
    if len(all_ranks) == 1:
        # only one rank, nothing to consolidate
        return reports

    if (style != ReportStyle.all and
            all(r["outcome"] == "passed" for r in reports)):
        # report from rank 0 if everything passed
        return [report for report in reports if report["_mpi_rank"] == 0]

    failed_ranks = set()
    for report in reports:
        rank = report["_mpi_rank"]
        # add rank to labels for ranks after 0, unless reporting all failures
        if rank > 0 or style in {ReportStyle.all, ReportStyle.all_failures}:
            report["nodeid"] = f"{nodeid} [rank={rank}]"
            report["location"][-1] = report["location"][-1] + f" [rank={rank}]"
        if report["outcome"] != "passed":
            failed_ranks.add(report["_mpi_rank"])
    failed_ranks = sorted(failed_ranks)

    if style == ReportStyle.all:
        return reports

    elif style == ReportStyle.all_failures:
        # select all reports on failed ranks
        return [r for r in reports if r["_mpi_rank"] in failed_ranks]

    elif style == ReportStyle.first_failure:
        # return just the first error
        first_failed_rank = failed_ranks[0]

        return [r for r in reports if r["_mpi_rank"] == first_failed_rank]
    elif style == ReportStyle.concise:
        # group by 'unique' reports
        reports_by_rank = {}
        for report in reports:
            reports_by_rank.setdefault(report["_mpi_rank"], []).append(report)
        _seen_keys = {}
        collected_reports = []
        for rank, rank_reports in reports_by_rank.items():
            rank_key = tuple(_report_key(report) for report in rank_reports)
            if rank_key in _seen_keys:
                _seen_keys[rank_key].append(rank)
            else:
                _seen_keys[rank_key] = [rank]
                collected_reports.extend(rank_reports)
        return collected_reports
    else:
        raise ValueError(f"Unhandled ReportStyle: {style}")

    return reports


def mpi_runtest(item):
    """Replacement for runtest

    Runs a single test with mpiexec
    """
    mpi_mark = item.get_closest_marker(MPI_MARKER_NAME)
    # allow parametrization
    if getattr(item, "callspec", None) and "mpiexec_n" in item.callspec.params:
        n = item.callspec.params["mpiexec_n"]
    else:
        n = mpi_mark.kwargs.get("n", 2)
    timeout = mpi_mark.kwargs.get("timeout", 120)
    exe = [
        MPIEXEC,
        "-n",
        str(n),
        sys.executable,
        "-m",
        "pytest",
        "--quiet",
        "--no-header",
        "--no-summary",
        f"{item.fspath}::{item.name}",
    ]
    env = dict(os.environ)
    env[MPI_SUBPROCESS_ENV] = "1"
    # add the mpiexec command for easy re-run
    item.add_report_section(
        "setup", "mpiexec command", f"{MPI_SUBPROCESS_ENV}=1 {shlex.join(exe)}"
    )

    with TemporaryDirectory() as reportlog_dir:
        env[TEST_REPORT_DIR_ENV] = reportlog_dir
        try:
            p = subprocess.run(
                exe,
                capture_output=True,
                text=True,
                env=env,
                timeout=timeout,
            )
        except subprocess.TimeoutExpired as e:
            if e.stdout:
                item.add_report_section(
                    "mpiexec pytest", "stdout",
                    e.stdout.decode("utf8", "replace")
                )
            if e.stderr:
                item.add_report_section(
                    "mpiexec pytest", "stderr",
                    e.stderr.decode("utf8", "replace")
                )
            pytest.fail(
                f"mpi test did not complete in {timeout} seconds",
                pytrace=False,
            )

        # Collect logs from all ranks
        reports = {}
        for rank in range(n):
            reportlog_file = os.path.join(reportlog_dir,
                                          f"reportlog-{rank}.jsonl")
            if os.path.exists(reportlog_file):
                with open(reportlog_file) as f:
                    for line in f:
                        report = json.loads(line)
                        if report["$report_type"] != "TestReport":
                            continue
                        report["_mpi_rank"] = rank
                        nodeid = report["nodeid"]
                        reports.setdefault(nodeid, []).append(report)

        for nodeid, report_list in reports.items():
            # consolidate reports according to config
            reports[nodeid] = consolidate_reports(
                nodeid, report_list, REPORT_STYLE)

        # collect report items for the test
        for report in chain(*reports.values()):
            if report["$report_type"] == "TestReport":
                # reconstruct and redisplay the report
                r = item.config.hook.pytest_report_from_serializable(
                    config=item.config, data=report
                )
                item.config.hook.pytest_runtest_logreport(
                    config=item.config, report=r)

    if p.returncode or not reports:
        if p.stdout:
            item.add_report_section("mpiexec pytest", "stdout", p.stdout)
        if p.stderr:
            item.add_report_section("mpiexec pytest", "stderr", p.stderr)
    if not reports:
        pytest.fail("No test reports captured from mpi subprocess!",
                    pytrace=False)
